#!/usr/bin/env python3
import os
from cnoid.Util import *
from cnoid.Base import *
from cnoid.Body import *
from cnoid.BodyPlugin import *
from cnoid.ROSPlugin import *
import math
import yaml
import csv
import roslib.packages
from rospy.timer import sleep

item_list = []
robot_item = []
env_item = []
sim_item = []
world_item = WorldItem()
yaml_file_path = ""
pkg_dir = ""
l = []

def readYaml():
    try:
        global item_list, pkg_dir, yaml_file_path
        file_nm = "cnoid_item_list.yaml"
        pkg_dir = roslib.packages.get_pkg_dir("cnoid_turtlebot_bringup")
        for dir_path, dirs, files in os.walk(pkg_dir + "/param"):
            for file in files:
                if file.endswith(file_nm):
                    pkg_dir = dir_path
                    yaml_file_path = dir_path + "/" + file
        # Read yaml file.
        with open(yaml_file_path, 'r') as f:
            item_list = yaml.safe_load(f)
    except Exception as e:
        print('[ERROR] Specified file is not found.')

def saveYaml():
    try:
        global item_list, yaml_file_path
        with open(yaml_file_path, "w") as f:
            yaml.dump(item_list, f)
    except Exception as e:
        print('[ERROR] Specified file is not found.')

def readCsv():
    try:
        global l, pkg_dir
        file_nm = "robot_state.csv"
        for dir_path, dirs, files in os.walk(pkg_dir + "/param"):
            for file in files:
                if file.endswith(file_nm):
                    file_path = dir_path + "/" + file
        with open(file_path) as f:
            reader = csv.reader(f)
            l = [ row for row in reader ]
    except Exception as e:
        print('[ERROR] Specified file is not found.')

def sceneViewProperties():
    # Set view properties.
    sceneWidget = SceneView.instance.sceneWidget
    sceneWidget.setHeadLightEnabled(True)
    sceneWidget.setHeadLightIntensity(0.1)
    sceneWidget.setFloorGridEnabled(False)
    sceneWidget.setWorldLightIntensity(0.25)
    sceneWidget.setWorldLightAmbient(0.25)
    sceneWidget.setBackgroundColor([0.1, 0.1, 0.3])
    sceneWidget.setCameraPosition(
        [ -15.6994, -1.70142, 23.3002 ],
        [ 0.571776, -0.00558784, -0.820391 ],
        [ 0.820352, -0.00801712, 0.571803 ])
    #    [ 0.5, 10.5, 1.4 ],
    #    [ -0.07, -0.98, -0.15 ],
    #    [ -0.01, -0.15, 0.98 ])

def worldItemGenerator():
    global world_item
    # Generate world item.
    RootItem.instance.addChildItem(world_item)

def readRobotModel():
    global item_list, robot_item
    
    if checkKey("ROS", item_list):
        # Read WorldROS item.
        readWorldRos()
    try:
        readBodyModel(item_list, robot_item, 'Robot')
    except Exception as e:
        print("[Error]", e)

def readBodyModel(item, body_item, e1 = ""):
    try:
        global world_item, l
        # Number of body models.
        body_num = len(item[e1])
        folder = []
        for i in range(body_num):
            if not checkKey("uri", item, e1, i):
                folder_item = FolderItem()
                folder_item.name = item[e1][i]['name']
                parent = item[e1][i]['parent']
                body_ret = checkParentKey(parent, body_item)
                folder_ret = checkParentKey(parent, folder)
                if parent == "World":
                    # folder is NULL
                    world_item.addChildItem(folder_item)
                    folder.append(folder_item)
                elif len(body_item) > 0 and body_ret[0]:
                    num = body_ret[1]
                    body_item[num].addChildItem(folder_item)
                    folder.append(folder_item)
                elif len(folder) > 0 and folder_ret[0]:
                    # Otherwise
                    num = folder_ret[1]
                    folder[num].addChildItem(folder_item)
                    folder.append(folder_item)
#                 body_item.append(folder_item)
            else:
                # Generate BodyItem.
                tmp = BodyItem()
                # Load body model.
                tmp.load(item[e1][i]['uri'])
                tmp.name = item[e1][i]['name']
                # Display robot model.
                tmp.setChecked(True)
                body = tmp.body
                if checkKey("translation", item, e1, i):
                    # Check existence of "translation" key.
                    trans = item[e1][i]['translation']
#                     for j in range(len(l)):
#                         if tmp.name in l[j]:
#                             trans = [ float(l[j][1]), float(l[j][2]), float(l[j][3]) ]
#                             item[e1][i]['translation'] = trans
#                         
                    body.rootLink.setTranslation(trans)
                else:
                    # Otherwise
                    body.rootLink.setTranslation([ 0.0, 0.0, 0.0 ])
                    
                if checkKey("rotation", item, e1, i):
                    # Check existence of "rotation" key.
                    r = math.radians(item[e1][i]['rotation'][0])
                    p = math.radians(item[e1][i]['rotation'][1])
                    y = math.radians(item[e1][i]['rotation'][2])

#                     for j in range(len(l)):
#                         if tmp.name in l[j]:
#                             r = math.radians(float(l[j][4]))
#                             p = math.radians(float(l[j][5]))
#                             y = math.radians(float(l[j][6]))
#                             item[e1][i]['rotation'] = [ float(l[j][4]), float(l[j][5]), float(l[j][6]) ]

                else:
                    # Otherwise
                    r = p = y = 0.0
                
                body.rootLink.setRotation(rotFromRpy([r, p, y]))
                body.calcForwardKinematics()
                tmp.storeInitialState()
                
                if not checkKey("parent", item, e1, i) or item_list[e1][i]['parent'] == "World":
                    # Check existence of "parent" key. Or "parent" key is "World"
                    world_item.addChildItem(tmp)
                else:
                    # Otherwise
                    # Parent item name.
                    parent = item_list[e1][i]['parent']
                    # Parent key existence check.
                    body_ret = checkParentKey(parent, body_item)
                    folder_ret = checkParentKey(parent, folder)
                    if len(body_item) > 0 and body_ret[0]:
                        # if it exists in the body item list.
                        num = body_ret[1]
                        body_item[num].addChildItem(tmp)
                    elif len(folder) > 0 and folder_ret[0]:
                        # if it exists in the folder item list.
                        num = folder_ret[1]
                        folder[num].addChildItem(tmp)
                    else:
                        # Otherwise
                        folder_item = FolderItem()
                        folder_item.name = item_list[e1][i]['parent']
                        world_item.addChildItem(folder_item)
                        folder_item.addChildItem(tmp)
                        folder.append(folder_item)
                # Add body item list.
                body_item.append(tmp)
        
                if checkKey("controller", item, e1, i):
                    # Check existence of "controller" key.
                    controller_item = SimpleControllerItem()
                    controller_item.name = item[e1][i]['controller']
                    controller_item.setController(controller_item.name)
                    if checkKey("controllerOption", item, e1, i):
                        # Check existence of "controllerOption" key.
                        controller_item.setOptions(item[e1][i]['controllerOption'])
                    body_item[i].addChildItem(controller_item)
    
                if e1 == "Robot" and checkKey("ROS", item):
                    # Read BodyROS item.
                    readBodyRos(body_item, i)
    except Exception as e:
        print("[ERROR]", e)

def checkKey(key, item, e1 = "", e2 = -1):
    try:
        if not e1:
            if not key in item:
                return False
        elif e1 and e2 == -1:
            if not key in item[e1]:
                return False
        else:
            if not key in item[e1][e2]:
                return False
        return True
    except Exception as e:
        print("[ERROR]", e)

def checkParentKey(parent, body_item):
    try:
        length = len(body_item)
        for i in range(length):
            if parent in body_item[i].name:
                return True, i
        return False, -1
    except Exception as e:
        print("[ERROR]", e)

def readWorldRos():
    global item_list, world_item
    
    try:
        worldRosItem = SubProjectItem()
        key = 'WorldROS'
        idx = 0
        if checkKey('uri', item_list, 'ROS', 0) and key in item_list['ROS'][0]['uri']:
            idx = 0
        elif checkKey('uri', item_list, 'ROS', 1) and key in item_list['ROS'][1]['uri']:
            idx = 1
            
        if not checkKey('name', item_list, 'ROS', idx):
            worldRosItem.name = "WorldROS"
        else:
            worldRosItem.name = item_list['ROS'][idx]['name']
        if not checkKey('uri', item_list, 'ROS', idx):
            print("[ERROR] Specified file is not found.")
        else:
            worldRosItem.load(item_list['ROS'][idx]['uri'])
            world_item.addChildItem(worldRosItem)
    except Exception as e:
        print("[Error]", e)

def readBodyRos(body_item, num):
    global item_list
    
    try:
        bodyRosItem = SubProjectItem()
        key = 'BodyROS'
        idx = 0
        if checkKey('uri', item_list, 'ROS', 0) and key in item_list['ROS'][0]['uri']:
            idx = 0
        elif checkKey('uri', item_list, 'ROS', 1) and key in item_list['ROS'][1]['uri']:
            idx = 1
            
        if not checkKey('name', item_list, 'ROS', idx):
            bodyRosItem.name = "BodyROS"
        else:
            bodyRosItem.name = item_list['ROS'][idx]['name']
        if not checkKey('uri', item_list, 'ROS', idx):
            print("[ERROR] Specified file is not found.")
        else:
            bodyRosItem.load(item_list['ROS'][idx]['uri'])
            body_item[num].addChildItem(bodyRosItem)
    except Exception as e:
        print("[Error]", e)

def readEnvironmentModel():
    global item_list, env_item
    
    try:
        readBodyModel(item_list, env_item, 'Environment')
    except Exception as e:
        print("[Error]", e)

def simulatorItemGenerator(item, e1 = ""):
    global sim_item, item_list
    try:
        if checkKey('type', item, e1, 0):
            if item[e1][0]['type'] == "AISTSimulator":
                simulator_item = AISTSimulatorItem()
                simulator_item.setTimeStep(0.001)
                simulator_item.setRealtimeSyncMode(True)
                simulator_item.setTimeRangeMode(SimulatorItem.TR_UNLIMITED)
                simulator_item.setMaxNumIterations(1000)
                simulator_item.setContactCorrectionVelocityRatio(30)
            elif item[e1][0]['type'] == "ODESimulator":
                simulator_item = SubProjectItem()
                simulator_item.name = item[e1][0]['name']
                if checkKey('uri', item, e1, 0):
                    simulator_item.load(item[e1][0]['uri'])
                print("ODESimulator")
            elif item[e1][0]['type'] == "AGXSimulator":
                simulator_item = SubProjectItem()
                simulator_item.name = item[e1][0]['name']
                if checkKey('uri', item, e1, 0):
                    simulator_item.load(item[e1][0]['uri'])
                print("AGXSimulator")
            else:
                print("[ERROR] Undefined simulator item.")
                return False
            world_item.addChildItem(simulator_item)
            simulator_item.setSelected(True)
            sim_item.append(simulator_item)
            # Generate sub simulator item
            subSimulatorItemGenerator(item_list, 'SubSimulator')
            return True
        else:
            return False
    except Exception as e:
        print("[ERROR]", e)
    
def subSimulatorItemGenerator(item, e1 = ""):
    global world_item
    try:
        subsim_num = len(item[e1])
        for i in range(subsim_num):
            if checkKey('type', item, e1, i) and item[e1][i]['type'] == "GLVisionSimulator":
                subSimulator_item = GLVisionSimulatorItem()
                subSimulator_item.name = item[e1][i]['name']
                if checkKey("target_body", item, e1, i) and item[e1][i]['target_body']:
                    subSimulator_item.setTargetBodies(item[e1][i]['target_body'])
                if checkKey("target_sensor", item, e1, i) and item[e1][i]['target_sensor']:
                    subSimulator_item.setTargetSensors(item[e1][i]['target_sensor'])
                if not sim_item[0].childItem:
                    simulator = sim_item[0]
                else:
                    simulator = sim_item[0].childItem
                simulator.addChildItem(subSimulator_item)
                    
    except Exception as e:
        print("[ERROR]", e)

if __name__ == "__main__":
    try:
        # Read yaml file.
        readYaml()
        # Read csv file.
        readCsv()
        # Generate world item.
        worldItemGenerator()
        # Generate simulator item
        simulatorItemGenerator(item_list, 'Simulator')
        # Read robot models.
        readRobotModel()
        # Read environment models.
        readEnvironmentModel()
#         # Generate simulator item
#         simulatorItemGenerator(item_list, 'Simulator')
        # Set scene view properties.
        sceneViewProperties()
        # Overwrite yaml file.
        saveYaml()
        # Start simulation.
        # sim_item[0].startSimulation()
    except Exception as e:
        print("[ERROR]", e)
